# frozen_string_literal: true

module Referator
  class Script
    attr_reader :workdir, :vars, :args, :env, :stdin

    def initialize(workdir:, vars:, args: [], env: {}, stdin: [])
      self.workdir = workdir
      self.vars = vars

      self.args = args
      self.env = env
      self.stdin = stdin

      raise 'Invalid script' if self.args.empty?
    end

    def call(**kwargs)
      raise 'Invalid vars' unless kwargs.keys.sort == vars

      stdout, status = Open3.capture2(
        sub_env(**kwargs),
        *sub_args(**kwargs),
        chdir: workdir.to_s,
        stdin_data: sub_stdin(**kwargs),
      )
      raise 'Exit code' unless status.success?

      String(stdout).freeze
    end

  private

    def workdir=(workdir)
      workdir = Pathname.new(workdir).expand_path.freeze
      raise 'Expected absolute path' unless workdir.absolute?
      raise 'Expected existing directory' unless workdir.directory?

      @workdir = workdir
    end

    def vars=(vars)
      @vars = [*vars].uniq.sort.freeze.each do |var|
        unless var.instance_of? Symbol
          raise TypeError, "Expected #{Symbol}, got #{var.class}"
        end
      end
    end

    def args=(args)
      @args = [*args].map { |arg| Sub.new vars, arg }.freeze
    end

    def env=(env)
      @env = Hash(env).to_h do |key, value|
        [String(key).freeze, Sub.new(vars, value)]
      end.freeze
    end

    def stdin=(stdin)
      @stdin = Sub.new vars, stdin
    end

    def sub_args(**kwargs)
      args.map { |sub| sub.call(**kwargs) }.freeze
    end

    def sub_env(**kwargs)
      env.transform_values { |sub| sub.call(**kwargs) }.freeze
    end

    def sub_stdin(**kwargs)
      stdin.call(**kwargs)
    end

    class Sub
      attr_reader :vars, :template

      def initialize(vars, template)
        self.vars = vars

        self.template =
          if template.instance_of?(String) ||
             template.instance_of?(Symbol) ||
             template.instance_of?(Hash)
            [template]
          else
            template
          end
      end

      def call(**kwargs)
        raise 'Invalid vars' unless kwargs.keys.sort == vars

        template.map do |part|
          if part.instance_of? String
            part
          else
            kwargs.fetch part
          end
        end.join.freeze
      end

    private

      def vars=(vars)
        @vars = [*vars].uniq.sort.freeze.each do |var|
          Referator.validate_script_var! var
        end
      end

      def template=(template)
        raise 'Invalid format' unless template.instance_of? Array

        @template = template.map do |part|
          part = String(part.keys.first).to_sym if part.instance_of? Hash

          unless part.instance_of?(String) ||
                 part.instance_of?(Symbol) && vars.include?(part)
            raise 'Invalid format'
          end

          part.freeze
        end.freeze
      end
    end
  end
end
